import pytest
import numpy as np
import spindry as spd


@pytest.fixture(
    params=(
        (spd.Atom(id=0, element_string='N'), 0, 'N', 1.4882656711484),
        (spd.Atom(id=65, element_string='P'), 65, 'P', 1.925513788),
        (spd.Atom(id=2, element_string='C'), 2, 'C', 1.60775485914852),
    )
)
def atom_info(request):
    return request.param


@pytest.fixture(
    params=(
        (spd.Bond(id=0, atom_ids=(0, 1)), 0, 0, 1),
        (spd.Bond(id=65, atom_ids=(2, 3)), 65, 2, 3),
        (spd.Bond(id=2, atom_ids=(3, 4)), 2, 3, 4),
        (spd.Bond(id=3, atom_ids=(0, 9)), 3, 0, 9),
    )
)
def bond_info(request):
    return request.param


@pytest.fixture
def atoms():
    return [spd.Atom(0, 'C'), spd.Atom(1, 'C')]


@pytest.fixture
def position_matrix():
    return np.array([[0, 0, 0], [0, 1.5, 0]])


@pytest.fixture
def position_matrix2():
    return np.array([[0, 0, 0], [0, 3, 0]])


@pytest.fixture
def displacement():
    return np.array([0, 1, 0])


@pytest.fixture
def displaced_position_matrix():
    return np.array([[0, 1, 0], [0, 2.5, 0]])


@pytest.fixture
def centroid():
    return np.array([0, 0.75, 0])


@pytest.fixture
def num_atoms():
    return 2


@pytest.fixture
def molecule(atoms, position_matrix):
    return spd.Molecule(
        atoms=atoms,
        bonds=[],
        position_matrix=position_matrix
    )


@pytest.fixture
def components(atoms, position_matrix):
    return (
        spd.Molecule(
            atoms=(atoms[0], ),
            bonds=[],
            position_matrix=position_matrix[0]
        ),
        spd.Molecule(
            atoms=(atoms[1], ),
            bonds=[],
            position_matrix=position_matrix[1]
        ),
    )


@pytest.fixture
def smolecule(atoms, position_matrix):
    return spd.SupraMolecule(
        atoms=atoms,
        bonds=[],
        position_matrix=position_matrix,
    )


@pytest.fixture
def distances():
    return np.array([1, 1.2, 1.4, 1.6, 1.8, 2.0, 4])


@pytest.fixture
def nonbond_potentials():
    return [
        29.650582241279984, 0.0, -1.1965106134640058,
        -0.7315108180046076, -0.40042074284821816,
        -0.22239608831999993, -0.0036423427949999992
    ]


@pytest.fixture
def nb_mins():
    return [1, 2, 4, 3]


@pytest.fixture
def final_potential():
    return -0.0564393681661353


@pytest.fixture
def final_pos_mat():
    return np.array([
        [-0.02514959, -0.50457139, -0.20141202],
        [ 1.31683193,  2.57903968,  0.21028248]
    ])


@pytest.fixture
def spinner():
    return spd.Spinner(
        step_size=0.5,
        rotation_step_size=5,
        num_conformers=50,
    )


@pytest.fixture
def spdpotential():
    return spd.SpdPotential()


@pytest.fixture
def radii_combinations():
    return [
        (1.24235230881914, 1.24235230881914, 1.24235230881914),
        (1.24235230881914, 2.056812828, 1.64958256840957),
        (1.75038091053539, 1.24235230881914, 1.496366609677265),
        (1.60775485914852, 1.4882656711484, 1.54801026514846),
    ]


@pytest.fixture
def final_potential2():
    return -0.11931353919461599


@pytest.fixture
def final_pos_mat2():
    return np.array([
        [0., 0., 0.],
        [0.55315151,  2.61708819, -1.322678]
    ])


@pytest.fixture
def movable_components():
    return (1, )



import pytest
import os
import numpy as np


def test_molecule_get_position_matrix(molecule, position_matrix):
    assert np.all(np.allclose(
        position_matrix,
        molecule.get_position_matrix(),
    ))


def test_molecule_with_position_matrix(molecule, position_matrix2):
    test = molecule.with_position_matrix(position_matrix2)
    assert np.all(np.allclose(
        position_matrix2,
        test.get_position_matrix(),
    ))


def test_molecule_with_displacement(
    molecule,
    displacement,
    displaced_position_matrix,
):
    test = molecule.with_displacement(displacement)
    assert np.all(np.allclose(
        displaced_position_matrix,
        test.get_position_matrix(),
    ))


@pytest.fixture
def path(request, tmpdir):
    return os.path.join(tmpdir, 'molecule.xyz')


def test_molecule_write_xyz_file(molecule, path):
    molecule.write_xyz_file(path)
    content = molecule._write_xyz_content()
    with open(path, 'r') as f:
        test_lines = f.readlines()

    assert ''.join(test_lines) == ''.join(content)


def test_molecule_get_atoms(molecule, atoms):
    for test, atom in zip(molecule.get_atoms(), atoms):
        assert test.get_id() == atom.get_id()
        assert test.get_element_string() == atom.get_element_string()


def test_molecule_get_centroid(
    molecule, position_matrix, centroid
):
    test = molecule.with_position_matrix(position_matrix)
    assert np.all(np.allclose(
        centroid,
        test.get_centroid(),
        atol=1E-6,
    ))


def test_molecule_get_num_atoms(molecule, num_atoms):
    assert molecule.get_num_atoms() == num_atoms



"""
SupraMolecule
=============

#. :class:`.SupraMolecule`

SupraMolecule class for optimisation.

"""

import networkx as nx
from .molecule import Molecule
import numpy as np


class SupraMolecule(Molecule):
    """
    Representation of a supramolecule containing atoms and positions.

    """

    def __init__(
        self,
        atoms,
        bonds,
        position_matrix,
        cid=None,
        potential=None,
    ):
        """
        Initialize a :class:`Supramolecule` instance.

        Parameters
        ----------
        atoms : :class:`iterable` of :class:`.Atom`
            Atoms that define the molecule.

        bonds : :class:`iterable` of :class:`.Bond`
            Bonds between atoms that define the molecule.

        position_matrix : :class:`numpy.ndarray`
            A ``(n, 3)`` matrix holding the position of every atom in
            the :class:`.Molecule`.

        cid : :class:`int`, optional
            Conformer id of supramolecule.

        potential : :class:`float`, optional
            Potential energy of Supramolecule.

        """

        self._atoms = tuple(atoms)
        self._bonds = tuple(bonds)
        self._position_matrix = np.array(
            position_matrix.T,
            dtype=np.float64,
        )
        self._define_components()
        self._cid = cid
        self._potential = potential

    @classmethod
    def init_from_components(cls, components):
        """
        Initialize a :class:`Supramolecule` instance from components.

        Parameters
        ----------
        components : :class:`iterable` of :class:`.Molecule`
            Molecular components that define the supramolecule.

        """

        atoms = []
        bonds = []
        position_matrix = []
        for comp in components:
            for a in comp.get_atoms():
                atoms.append(a)
            for b in comp.get_bonds():
                bonds.append(b)
            for pos in comp.get_position_matrix():
                position_matrix.append(pos)

        supramolecule = cls.__new__(cls)
        supramolecule._atoms = tuple(atoms)
        supramolecule._bonds = tuple(bonds)
        supramolecule._components = tuple(components)
        supramolecule._cid = None
        supramolecule._potential = None
        supramolecule._position_matrix = np.array(position_matrix).T
        return supramolecule

    def _define_components(self):
        """
        Define disconnected component molecules as :class:`.Molecule`s.

        """

        # Produce a graph from the molecule that does not include edges
        # where the bonds to be optimized are.
        mol_graph = nx.Graph()
        for atom in self.get_atoms():
            mol_graph.add_node(atom.get_id())

        # Add edges.
        for bond in self._bonds:
            pair_ids = (bond.get_atom1_id(), bond.get_atom2_id())
            mol_graph.add_edge(*pair_ids)

        # Get atom ids in disconnected subgraphs.
        comps = []
        for c in nx.connected_components(mol_graph):
            in_atoms = [
                i for i in self._atoms
                if i.get_id() in c
            ]
            in_bonds = [
                i for i in self._bonds
                if i.get_atom1_id() in c and i.get_atom2_id() in c
            ]
            new_pos_matrix = self._position_matrix[:, list(c)].T
            comps.append(
                Molecule(in_atoms, in_bonds, new_pos_matrix)
            )

        self._components = tuple(comps)

    def _write_xyz_content(self):
        """
        Write basic `.xyz` file content of Molecule.

        """
        coords = self.get_position_matrix()
        content = [0]
        for i, atom in enumerate(self.get_atoms(), 1):
            x, y, z = (i for i in coords[atom.get_id()])
            content.append(
                f'{atom.get_element_string()} {x:f} {y:f} {z:f}\n'
            )
        # Set first line to the atom_count.
        content[0] = f'{i}\ncid:{self._cid}, pot: {self._potential}\n'

        return content

    def get_components(self):
        """
        Yields each molecular component.

        """

        for i in self._components:
            yield i

    def get_cid(self):
        return self._cid

    def get_potential(self):
        return self._potential

    def __str__(self):
        return repr(self)

    def __repr__(self):
        comps = ', '.join([str(i) for i in self.get_components()])
        return (
            f'{self.__class__.__name__}('
            f'{len(list(self.get_components()))} components, '
            f'{comps})'
        )



import numpy as np


def test_opt(spinner, smolecule, final_pos_mat, final_potential):
    test = spinner.get_final_conformer(smolecule)
    print(test.get_position_matrix())
    assert np.all(np.allclose(
        final_pos_mat,
        test.get_position_matrix(),
    ))
    assert np.isclose(
        spinner._compute_potential(test),
        final_potential
    )


def test_opt_test_move(spinner):
    # Do not test random component.
    assert spinner._test_move(curr_pot=-1, new_pot=-2)


def test_opt_setcomp(
    spinner,
    smolecule,
    final_pos_mat2,
    final_potential2,
    movable_components,
):
    test = spinner.get_final_conformer(smolecule, movable_components)
    print(test.get_position_matrix())
    assert np.all(np.allclose(
        final_pos_mat2,
        test.get_position_matrix(),
    ))
    assert np.isclose(
        spinner._compute_potential(test),
        final_potential2
    )



"""
Bond
====

#. :class:`.Bond`

Bond class.

"""


class Bond:
    """
    Bond between two atoms.

    """

    def __init__(self, id, atom_ids):
        """
        Initialize a :class:`Bond` instance.

        Parameters
        ----------
        id : :class:`int`
            ID to be assigned to bond.

        atom_ids : :class:`iterable` of :class:`int`
            IDs of atom 1 and atom 2 in bond, where atom 1 is always
            the smaller number and the IDs cannot be the same.

        """

        self._id = id
        if len(set(atom_ids)) == 0:
            raise ValueError('Two distict atom ids are required.')
        self._atom1_id, self._atom2_id = sorted(atom_ids)

    def get_id(self):
        """
        Get bond ID.

        """

        return self._id

    def get_atom1_id(self):
        """
        Get ID of atom 1 in bond.

        """

        return self._atom1_id

    def get_atom2_id(self):
        """
        Get ID of atom 2 in bond.

        #"""

        return self._atom2_id

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(id={self.get_id()}, '
            f'atom1_id={self.get_atom1_id()}, '
            f'atom2_id={self.get_atom2_id()})'
        )



"""
Potential
=========

#. :class:`.Potential`

Classes for calculating the potential energy of supramolecules.

"""

import numpy as np

from itertools import combinations
from scipy.spatial.distance import cdist


class Potential:
    """
    Base class for potential calculators.

    """

    def __init__(self):
        """
        Initialize a :class:`Spinner` instance.

        """

    def compute_potential(self, supramolecule):
        """
        Calculate potential energy.

        Parameters
        ----------
        supramolecular : :class:`spd.Supramolecule`
            Supramolecule to evaluate.

        """
        raise NotImplementedError()



class SpdPotential:
    """
    Default spindry non-bonded potential function.

    """

    def __init__(self, nonbond_epsilon=5):
        """
        Initialize a :class:`Spinner` instance.

        Parameters
        ----------
        nonbond_epsilon : :class:`float`, optional
            Value of epsilon used in the nonbond potential in MC moves.
            Determines strength of the nonbond potential.
            Defaults to 20.

        """

        self._nonbond_epsilon = nonbond_epsilon

    def _nonbond_potential(self, distance, sigmas):
        """
        Define a Lennard-Jones nonbonded potential.

        This potential has no relation to an empircal forcefield.

        """

        return (
            self._nonbond_epsilon * (
                (sigmas/distance) ** 12 - (sigmas/distance) ** 6
            )
        )

    def _mixing_function(self, val1, val2):
        return (val1 + val2) / 2

    def _combine_sigma(self, radii1, radii2):
        """
        Combine radii using Lorentz-Berthelot rules.

        """

        len1 = len(radii1)
        len2 = len(radii2)

        mixed = np.zeros((len1, len2))
        for i in range(len1):
            for j in range(len2):
                mixed[i, j] = self._mixing_function(
                    radii1[i], radii2[j],
                )

        return mixed

    def _compute_nonbonded_potential(self, position_matrices, radii):
        nonbonded_potential = 0
        for pos_mat_pair, radii_pair in zip(
            combinations(position_matrices, 2),
            combinations(radii, 2),
        ):
            pair_dists = cdist(pos_mat_pair[0], pos_mat_pair[1])
            sigmas =  self._combine_sigma(radii_pair[0], radii_pair[1])
            nonbonded_potential += np.sum(
                self._nonbond_potential(
                    distance=pair_dists.flatten(),
                    sigmas=sigmas.flatten(),
                )
            )

        return nonbonded_potential

    def compute_potential(self, supramolecule):
        component_position_matrices = (
            i.get_position_matrix()
            for i in supramolecule.get_components()
        )
        component_radii = (
            tuple(j.get_radius() for j in i.get_atoms())
            for i in supramolecule.get_components()
        )
        return self._compute_nonbonded_potential(
            position_matrices=component_position_matrices,
            radii=component_radii,
        )



"""
This module defines general-purpose objects, functions and classes.

"""

from scipy.spatial.transform import Rotation
from scipy.spatial.distance import euclidean
import numpy as np


def get_atom_distance(position_matrix, atom1_id, atom2_id):
    """
    Return the distance between two atoms.

    """

    return float(euclidean(
        u=position_matrix[atom1_id],
        v=position_matrix[atom2_id]
    ))


def rotation_matrix_arbitrary_axis(angle, axis):
    """
    Returns a rotation matrix of `angle` radians about `axis`.
    Parameters
    ----------
    angle : :class:`float`
        The size of the rotation in radians.
    axis : :class:`numpy.ndarray`
        A 3 element aray which represents a vector. The vector is the
        axis about which the rotation is carried out. Must be of
        unit magnitude.
    Returns
    -------
    :class:`numpy.ndarray`
        A ``3x3`` array representing a rotation matrix.
    """

    a = np.cos(angle/2)
    b, c, d = axis * np.sin(angle/2)

    e11 = np.square(a) + np.square(b) - np.square(c) - np.square(d)
    e12 = 2*(b*c - a*d)
    e13 = 2*(b*d + a*c)

    e21 = 2*(b*c + a*d)
    e22 = np.square(a) + np.square(c) - np.square(b) - np.square(d)
    e23 = 2*(c*d - a*b)

    e31 = 2*(b*d - a*c)
    e32 = 2*(c*d + a*b)
    e33 = np.square(a) + np.square(d) - np.square(b) - np.square(c)

    # Initialize as a scipy Rotation object, which normalizes the
    # matrix and allows for returns as quaternion or alternative
    # type in the future.
    return Rotation.from_matrix(np.array([
        [e11, e12, e13],
        [e21, e22, e23],
        [e31, e32, e33]
    ])).as_matrix()



from .spinner import *  # noqa
from .molecule import *  # noqa
from .supramolecule import *  # noqa
from .atom import *  # noqa
from .bond import *  # noqa
from .utilities import *  # noqa
from .radii import *  # noqa
from .potential import *  # noqa



"""
Molecule
========

#. :class:`.Molecule`

Molecule class for optimisation.

"""

import numpy as np


class Molecule:
    """
    Representation of a molecule containing atoms and positions.

    """

    def __init__(self, atoms, bonds, position_matrix):
        """
        Initialize a :class:`Molecule` instance.

        Parameters
        ----------
        atoms : :class:`iterable` of :class:`.Atom`
            Atoms that define the molecule.

        bonds : :class:`iterable` of :class:`.Bond`
            Bonds between atoms that define the molecule.

        position_matrix : :class:`numpy.ndarray`
            A ``(n, 3)`` matrix holding the position of every atom in
            the :class:`.Molecule`.

        """

        self._atoms = tuple(atoms)
        self._bonds = tuple(bonds)
        self._position_matrix = np.array(
            position_matrix.T,
            dtype=np.float64,
        )

    def get_position_matrix(self):
        """
        Return a matrix holding the atomic positions.

        Returns
        -------
        :class:`numpy.ndarray`
            The array has the shape ``(n, 3)``. Each row holds the
            x, y and z coordinates of an atom.

        """

        return np.array(self._position_matrix.T)

    def with_displacement(self, displacement):
        """
        Return a displaced clone Molecule.

        Parameters
        ----------
        displacement : :class:`numpy.ndarray`
            The displacement vector to be applied.

        """

        new_position_matrix = (
            self._position_matrix.T + displacement
        )
        return Molecule(
            atoms=self._atoms,
            bonds=self._bonds,
            position_matrix=np.array(new_position_matrix),
        )

    def with_position_matrix(self, position_matrix):
        """
        Return clone Molecule with new position matrix.

        Parameters
        ----------
        position_matrix : :class:`numpy.ndarray`
            A position matrix of the clone. The shape of the matrix
            is ``(n, 3)``.

        """

        return Molecule(
            atoms=self._atoms,
            bonds=self._bonds,
            position_matrix=np.array(position_matrix),
        )

    def _write_xyz_content(self):
        """
        Write basic `.xyz` file content of Molecule.

        """
        coords = self.get_position_matrix()
        content = [0]
        for i, atom in enumerate(self.get_atoms(), 1):
            x, y, z = (i for i in coords[atom.get_id()])
            content.append(
                f'{atom.get_element_string()} {x:f} {y:f} {z:f}\n'
            )
        # Set first line to the atom_count.
        content[0] = f'{i}\n\n'

        return content

    def write_xyz_file(self, path):
        """
        Write basic `.xyz` file of Molecule to `path`.

        Connectivity is not maintained in this file type!

        """

        content = self._write_xyz_content()

        with open(path, 'w') as f:
            f.write(''.join(content))

    def get_atoms(self):
        """
        Yield the atoms in the molecule, ordered as input.

        Yields
        ------
        :class:`.Atom`
            An atom in the molecule.

        """

        for atom in self._atoms:
            yield atom

    def get_bonds(self):
        """
        Yield the bonds in the molecule, ordered as input.

        Yields
        ------
        :class:`.Bond`
            A bond in the molecule.

        """

        for bond in self._bonds:
            yield bond

    def get_num_atoms(self):
        """
        Return the number of atoms in the molecule.

        """

        return len(self._atoms)

    def get_centroid(self, atom_ids=None):
        """
        Return the centroid.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The ids of atoms which are used to calculate the
            centroid. Can be a single :class:`int`, if a single
            atom is to be used, or ``None`` if all atoms are to be
            used.

        Returns
        -------
        :class:`numpy.ndarray`
            The centroid of atoms specified by `atom_ids`.

        Raises
        ------
        :class:`ValueError`
            If `atom_ids` has a length of ``0``.

        """

        if atom_ids is None:
            atom_ids = range(len(self._atoms))
        elif isinstance(atom_ids, int):
            atom_ids = (atom_ids, )
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        if len(atom_ids) == 0:
            raise ValueError('atom_ids was of length 0.')

        return np.divide(
            self._position_matrix[:, atom_ids].sum(axis=1),
            len(atom_ids)
        )

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return (
            f'<{self.__class__.__name__}({len(self._atoms)} atoms) '
            f'at {id(self)}>'
        )
